/*
        Yapp 4 ZaCx!

        Versione 1.4 del 20/5/92

        (c) 1989-1992 Marco Savegnago IW3FQG

        Questo modulo contiene un prototipo di una procedura per trasferire
        files in protocollo Yapp secondo le specifiche di WA7MBL e le varie
        estensioni apportate in seguito da altri.

        In questa versione e' supportato il resume secondo il metodo proposto
        da FC1EBN e usato in TPK 170 completo di checksum.

        L'output sulla seriale avviene mediante lo stream caux che e' un
        driver per le seriali homemade.

        In questo stream ho usato l'overloading degli operatori >> e <<
        con lo stesso significato ( caux >> = input, caux << output) di quelli
        usati per lo stdio cout e cin.

        Rispetto a Yapp, ZaCx (quando il trasferimento avviene tra due ZaCx)
        riporta la data e l'ora di creazione e gli attributi del file.
        Questo avviene inserendo nel campo opzionale dell'header due stringhe
        contenenti la data e ora di creazione e gli attributi.
        L'header Yapp in trasferimenti originati da ZaCx ha la seguente
        forma:

        <SOH> <len> (Filename) <NUL> (File Size in ASCII) <NUL> \
                                     (Filedatetime) <NUL> (Fileattribute) <NUL>

	Il campo Filedatetime e' espesso mediante una stringa hex
	contenete un numero a 32 bit che al suo interno e' ripartito
	nei seguenti campi:

		 31 30 29 28 27 26 25 24 23 22 21 20 19 18 17 16
		[<---- anno-1980 --->|<---mese-->|<---giorno--->]

		 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
		[<----ore----->|<-----minuti----->|<-secondi/2->]

	che corrispondono rispettivamente al contenuto dei registri
	CX e DX al ritorno della funzione DOS 57h.
        (disponibili nella funzione getftime() in TurboC )

	Il campo Fileattribute e' espresso con una stringa hex a 8 bit
	il cui significato e' il seguente:

		Bit             Attributi
		---             ---------
		0               File di sola lettura
		1               File nascosto
		2               File di sistema
		5               File di archivio

	che corrispondono al contenuto del reg. CX nella funzione 43h.
        (disponibili nella funzione setftime() in TurboC )

        Nel caso che il corrispondondente stia usando una copia di ZaCx,
	il programma esegue la copia completa del file.
	Nel caso che il corrispondente non fornisca i parametri opzionali
	al file viene asseggnata la data e l'ora corrente di creazione e
	gli attributi standard.

	Nel Resume vengono controllati oltre al nome e la dimensione del
	file anche la date e ora di creazione e gli attributi.
        Nel caso di file con nome e lunghezza uguale ma con attributi o data
        differenti oppure con lunghezza minore di quella gia' ricevuta viene
        creato un nuovo file con estensione progressiva .000 .001 .002 etc.
	Questo viene segnalato a fianco del nome del file.
        Questo serve per evitare che eventuali file presenti nella directory
        di destinazione possano venir rovinati.

	Nel caso in cui il filedatetime e gli attributi non vengono
	trasmessi (Es: Yapp) non vengono presi in considerazione e nel
	confronto si tiene solo conto del nome e della lunghezza del
	file.
	Quest'ultima opzione viene mantenuta per motivi di compatibilita'.

        Salve, il motivo percui con la versione 5.14 del BBS di F6FBB i vari
        software in circolazione che supportano lo Yapp con il resume non
        funzionano piu' (tranne il TPK ma non so da quale versione) e' dovuto
        a una piccola modifica del protocollo di resume che aggiunge in coda a
        ogni frame Data Yapp un byte di checksum.
        Questo causa un errore che porta al blocco del trasferimento dopo
        aver ricevuto i primi 256 byte dopo la ripresa in qualsiasi programma
        che implementa il protocollo Yapp con il Resume fatto alla vecchia
        maniera (quello introdotto da FC1EBN in TPK (anche qui .... non mi
        ricordo la versione)).

        In pratica durante il resume uno Yapp Send_Data frame e' strutturato
        come segue:

        STX len (Data) chksum

        Rispetto a prima c'e' in piu' solo il byte chksum che NON va' contato
        come facente parte di (Data) e quindi e' escluso dal conteggio posto
        in len.
        Ripeto perche' forse non e' chiaro...
        il byte len contiene il numero di caratteri che seguono e che compongono
        la sequenza Data, chksum e' escluso da questo conteggio.
        Resta poi sempre valido il discorso che se len = 0 allora Data = 256.

        Questo checksum penso lo abbiano inserito per evitare che eventuali
        disturbi nel cavo seriale tra TNC<->Computer o perdite di caratteri
        usando ad esempio l'interfaccia TNC2 del G8BPQ (ma capita ancora??))
        possano rendere inutile il trasferimento (boh!).

        Leggete le modifiche direttamente nel file sorgente....

        Nel sorgente che segue sono comprese tutte le funzioni indispensabili
        al protocollo e utili per un programma di terminale DUMB come ZaCc.
        La gestione del timer e dell'hostmode di DRSI, TheFirmware e AEA di ZaCh
        non sono inclusi.

        Buon lavoro!

        Vicenza, 23/5/1992 Marco IW3FQG.

*/


#include <stdio.h>
#include <dos.h>
#include <string.h>
#include <stdlib.h>
#include <dir.h>
#include <io.h>
#include <bios.h>
#include <ctype.h>
#include <time.h>
#include <sys\stat.h>

#define NIL  '\0'
#define SOH  '\x01'
#define STX  '\x02'
#define ETX  '\x03'
#define EOT  '\x04'
#define ENQ  '\x05'
#define ACK  '\x06'
#define DLE  '\x10'
#define NAK  '\x15'
#define CAN  '\x18'

#define PUT_TRANS        0
#define PUT_CONV         1
#define T_START          2
#define T_STOP           3
#define S_HEADER         4
#define R_HEADER         5
#define T_FILE           6
#define R_FILE           7
#define FILE_T           8
#define FILE_R           9
#define DEST_STOP        10
#define ORG_STOP         11
#define YAPP_TIMEOUT     12
#define PROTOCOL_ERROR   13
#define YAPP_IO_ERROR    14
#define FILE_RESUMED     15

typedef enum {  S,S1,SH,SD,SE,ST,R,RH,RD,YAPP_ABORT,CW,C,YAPP_START }
								yapp_states;

typedef enum {  UK,RR,RF,SI,HD,DT,EF,ET,NR,CN,CA,RI,TX,UU,TM,AF,AT }
								packet_type;

/* Funzioni locali */

static void file_xfer(void);
static void SendInit (void);
static void SendInit_Retry (void);
static void Sendhdr (void);
static void Senddata (void);
static void SendEOF (void);
static void SendEOT (void);
static void Receive (void);
static void RcvHdr (void);
static void RcvData (void);
static void Cancel (void);
static void CanWait (void);
static void CanRecd (void);
static void Dummy (void);
static void showstate (void);
static int  waitready (void);
static void yapp_message (int num);
static void disppacket (void);
static void showbyte (void);
static void showheader(void);
static void split_header(void);
static void show_start_time(void);
static void getpkstr (void);
static void SendMSG (char *buf);
static void Trans (void);
static void Conv (void);
static void change_date_attr(void);
static void cambia_nome (void);

/*
   Descrizione dei vari stati Yapp
   Servono per stamparli nella window Yapp.
*/

static char *state_desc[] = {   "SendInit  ",
				"SendInit  ",
				"SendHeader",
				"SendData  ",
				"SendEof   ",
				"SendEOT   ",
				"RcvWait   ",
				"RcvHeader ",
				"RcvData   ",
				"SndABORT  ",
				"WaitAbtAck",
				"RcdABORT  ",
				"Start     "};

// Vettore con i puntantori alle varie procedure

static void (*state_funct[])(void) = {
                                        SendInit,
                                        SendInit_Retry,
                                        Sendhdr,
                                        Senddata,
                                        SendEOF,
                                        SendEOT,
                                        Receive,
                                        RcvHdr,
                                        RcvData,
                                        Cancel,
                                        CanWait,
                                        CanRecd,
                                        Dummy
                                      };


static int  sendinit_count;             // Contatore dei retry di SI
static long xfercnt;                    // Contatore dei byte trasferiti
static char xferok;                     // Flag per sapere se il trasferimento
					// ha avuto successo

static yapp_states state;               // stato corrente del protocollo
static packet_type ptype;               // tipo corrente di frame

static char pkbuff[256];                // Buffer dei frame Yapp
static int pklen;                       // Lunghezza del frame Yapp


static char yapp_drive[MAXDRIVE],       // Dest. Yapp drive
	    yapp_path[MAXPATH],         // Dest. Yapp path
	    full_yapp_filename[MAXPATH];// Nome completo del file

static FILE *Yapp_File;                 // Puntatore al file trasferito

static char auto_start_download;        // Flag per autodownload
static char simple_download;            // Flag per download semplice

/*
   Per download semplice intendo quello in cui la stazione che riceve il
  file specifica i nome del file che vuole ricevere.
*/

static struct ffblk gruppo_file;        // Struttura di controllo delle
					// chiamate a findfirst e findnext
					// nel caso di trasferimento di
					// piu' di un file

static char     primo_multi_file,       // Flag di controllo per trasferimenti
		altri_multi_file;       // multifile (wildcard).

static char     primo_select_file,      // Flag di controllo per trasferimenti
                altri_select_file;      // selezionati con zadir di IW3EZU

static char select_mode;                // Flag che seleziona modo select
					// (con ZADIR) o multi

static LISTA *lista_select_file;        // Puntatore alla lista di file passata
					// da zadir

static char filename[MAXFILE+MAXEXT],   // Componenti dell'header Yapp
            lunghezza[10],              // Lunghezze provvisorie
	    dataora[20],
	    accesso[10];

static char lunghezza_parziale[10];     // Stringa che contiene la lunghezza
					// recuperata in un resume
static char file_resumed;               // Flag di resume

static unsigned long byte_file_resumed; // Lunghezza del file recupreato

static char     file_date[20],          // Componenti della maschera
		file_time[20],
		tempo_parziale[10],
		cps[10];

static time_t   inizio,                 // Tempi utili nel trasferimento
		differenza_tempo;

static struct ftime Yapp_filetime;      // Filedatetime del file corrente
static char Yapp_filemode;              // Fileattribute del file corrente

static char *HaveThatFile = " Errore il file esiste gia'!\r";
static char *HavntThatFile= "?? Qui non c'e' niente con quel nome!  \r";

// Alcune variabili dipendenti dal programma in cui il modulo e' usato

static char Resume,resume_with_chksum;
static char Tnc_type;

static void file_xfer(void)
{
  xfercnt  = 0L;
  Resume   = ZaCc_Flag.yapp_resume;
  resume_with_chksum = ZaCc_Flag.resume_with_chksum;
  Tnc_type = ZaCc_Flag.tnc_type;
  sendinit_count = 0;
  file_resumed = NO;

  // Guarda che ore sono all'inizio del trasferimento

  time (&inizio);
 
  // se e' attivato lo strip dell'ottavo bit disabilitalo

  if (ZaCc_Flag.bit_strip) caux.mask('\xff');

  yapp_mask();                  // Visualizza la finestra di trasferimento
  yapp_message(PUT_TRANS);
  cmdmode();                    // Mette il TNC in modo cmd
  caux.flush();                 // Flussha l'Rx e il Tx buffer
  delay (500);
  Trans();                      // Mette il TNC in Transparent Mode
  delay (500);                  // Aspetta un altro po'
  caux.flush();                 // Flussha Rx e TX

  // I valori di delay (espressi in millisecondi) potrebbero essere diminuiti

  // E ora entra nel loop del protocollo finche' il trasferimento non
  // termina.
  // Le varie funzioni sono richiamate mediante i puntatori contenuti del
  // vettore state_funct[];

  do
    {
	showstate();
	(*state_funct[state])();
    }
  while (state != YAPP_START);


  delay (1000);                 // Apetta un po' che si veda che e' finito
  yapp_message(PUT_CONV);
  cmdmode();                    // Ed esegui le cose opposte di prima
  caux.flush();
  delay (500);
  Conv();
  delay (500);
  caux.flush();

  /* se e' attivato lo strip dell'ottavo bit riabilitalo */

  if (ZaCc_Flag.bit_strip) caux.mask('\x7f');
}

/*
   Inzia il trasferimento di uno o piu' file (specificati con i wildcard)
   richiedendo l'inserimento
   In caso di errore o file non trovato ritorna senza errori.
*/

void yapp_upload(void)
{
  if(!chiedi_nome(full_yapp_filename," [ YAPP UPLOAD ]")) return;

  fnsplit(full_yapp_filename,yapp_drive,yapp_path,NULL,NULL);
  if (!findfirst(full_yapp_filename,&gruppo_file,0))
			{ primo_multi_file=YES; state = S; file_xfer(); }

  primo_multi_file=NO; altri_multi_file=NO;
}




// Inizia un trasferimento prendendo i nomi dalla lista passata da zadir

void yapp_upload_selected(LISTA *lst)
{
  lista_select_file=lst;
  select_mode=YES; primo_select_file=YES;
  state = S;
  file_xfer();
  delete(lista_select_file);
  select_mode=NO; primo_select_file=NO;
}
/*
        Questa funzione viene richiamata dal filtro di caratteri di ZaCc
        quando si e' connessi con qualcuno e viene ricevuta la sequenza
        Yapp RI
*/

void yapp_auto_upload(void)
{
  // Viene dapprima prelevato il nome del file completo di path
  getpkstr();

  // Viene estratto il path dal nome
  fnsplit(pkbuff,yapp_drive,yapp_path,NULL,NULL);

  // E quindi viene ricercato il o i file (potrebbe esseci un wildcard)
  if (!findfirst(pkbuff,&gruppo_file,0))
  {
        // Se c'e' qualcosa trasmettilo
        primo_multi_file=YES;
        state = S;
        file_xfer();
  }
  // Altrimenti digli che non hai niente e ritorna
  else  caux << NAK << strlen(HavntThatFile) << HavntThatFile;
}

/*
   Download di un file con la richiesta del nome.
   Il nome nell'header non viene preso in considerazione.
*/

void yapp_download(void)
{
  if(!chiedi_nome(full_yapp_filename," [ YAPP DOWNLOAD ]")) return;

  if((Yapp_File = fopen(full_yapp_filename,"wb+")) == NULL)
					{ fclose(Yapp_File); return; }
  // Download con nome cambiato dal ricevente
  simple_download=YES;
  state = R;
  xferok = FALSE;
  file_xfer();

  if (!xferok){
	fclose(Yapp_File);
        if (!Resume)     // Cancella il file se Resume=0
		     remove(full_yapp_filename);
	// Nota che qui la date e l'ora di creazione e i file attr non vengono
        // modificati perche' il file puo' non avere lo stesso nome.
  }
  simple_download=NO;
}

/*
	Invia richiesta di trasmissione di un file seguendo spec Yapp 1.1
	Wildcard ammessi
*/


void Send_Receive_Init(void)
{
  if(!chiedi_nome(full_yapp_filename," [ REQUEST YAPP DOWNLOAD ]"))
								return;
  cmdmode();
  caux.flush();
  delay (500);
  Trans();
  delay (500);
  caux.flush();

  // Invia stringa con RI

  caux << ENQ << '\x02' << (char) strlen (full_yapp_filename) << full_yapp_filename;

  delay (2000);
  cmdmode();
  caux.flush();
  delay (500);
  Conv();
  delay (500);
  caux.flush();
}

/*
	Questa funzione fa' partire un download automaticamente.
	Viene richiamata direttamente dal filtro dei caratteri in ingresso
        di ZaCc quando si e' connessi e passa SI.
*/

void yapp_auto_download(void)
{
  auto_start_download=YES;
  state=RH;
  file_xfer();
  if (!xferok)
  {
	fclose(Yapp_File);
	if (!Resume) remove(full_yapp_filename);
	else change_date_attr();
  }
}

/*
	Questa funzione rimane in attesa del prossimo carattere rispettando
	i timeout come nello Yapp.
	Negli stati S (SendInit) o S1 (SendInit_Retry) il timeout e' di
	20 secondi (364 ticks (18.2 al secondo)), mentre per gli altri stati
	e' di 120 secondi.
*/

static int waitready (void)
{

  static int i=0;
  timeout timeout;
  int Ch;

  if (caux.isAny())
                return(FALSE);   // Se c'e' qualcosa in RX torna subito

  if ((state==S) || (state==S1))
                        timeout.setTimer(364);  // setta il timeout a 20sec.

  else  timeout.setTimer(2184);      // Altrimenti a 120 sec.

  // Finche' il timeout e' diverso da 0 e c'e' il DCD (Tnc connesso)

  while(!timeout.isTimeout() && caux.cd())
  {
        // Se c'e' qualcosa in rx torna
        if (caux.isAny()) return(FALSE);

        // Stampa l'ora e cps
        showbyte();
  }
  // Se invece il timer finisce a zero..
  ptype = TM;     // Timeout !!!!!
  return (TRUE);
}

static void Dummy (void) { /* NOOP */ }


/*
	Questa funzione trasmette un frame Yapp
        che vale 256 e potrebbe variare insieme ai timeout durante i
	trasferimenti in modo da adattarsi al traffico.
*/

static void getpkstr (void)
{
    // Vede se c'e' qualcosa

    if (waitready()) return;

    // Se si lo preleva

    caux >> pklen; pklen &= 0xff;

    // Se e' un frame Data e la lunghezza == 0 allora la lunghezza e'
    // di 256 caratteri

    if ((ptype == DT) && (pklen == 0))
                                pklen = 256;

    // Se invece la lunghezza e' == 0 ritorna

    if (!pklen) return;

    // Preleva uno a uno i byte (tenendo conto degli eventuali timeout)

    char chksum=0,chk=0;

    for (int i = 0; i < pklen; i++)
    {
	 if (waitready()) return;
         caux >> *(pkbuff+i);

         // Calcola il checksum per YappC

         chksum+=*(pkbuff+i);
    }

    if (ptype == DT && file_resumed && resume_with_chksum)
    {
	if (waitready()) return;

        // Preleva il checksum e lo compara con quello calcolato

        caux >> chk;

	if (chk != chksum)
	{
		ptype=CA;
		state=YAPP_ABORT;
	}
    }
}

/*
	Questa funzione preleva dal buffer l'inizio del prossimo frame e
	seleziona cosa fare.
*/

static void Getpack (void)
{
	char Ch;

	ptype = UK;
	if (waitready()) return;

        caux >> Ch;
	switch(Ch)
	{
		case ACK:
			if (waitready()) return;
                        caux >> Ch;
			switch (Ch)
			{
				case 0x01:
					ptype=RR;
					break;
				case 0x02:
					ptype=RF;
					break;
				case 0x03:
					ptype=AF;
					break;
				case 0x04:
					ptype=AT;
					break;
				case 0x05:
					ptype=CA;
					break;
				default:
					break;
			}
			break;      // fine ACK
		case ENQ:
			if (waitready()) return;
                        caux >> Ch;
			switch (Ch)
			{
				case 0x01:
					ptype=SI;
					break;
				case 0x02:
					ptype=RI;
					break;
				default:
					ptype=UU;
					break;
			}
			break;  // ENQ
		case SOH:
			ptype=HD;
			getpkstr();
			break;
		case STX:
			ptype=DT;
			getpkstr();
			break;
		case ETX:
			if (waitready()) return;
                        caux >> Ch;
			if (Ch == '\x01') ptype = EF;
			break;
		case EOT:
			if (waitready()) return;
                        caux >> Ch;
			if (Ch == '\x01') ptype = ET;
			break;
		case NAK:
			ptype=NR;
			getpkstr();
			break;
		case CAN:
			ptype=CN;
			getpkstr();
			break;
		case DLE:
			ptype=TX;
			getpkstr();
			break;
		default:
			break;
	}
}

/*
	Trasmette la sequenza SendInit e attende risposte dall'alta parte...
*/


static void SendInit(void)
{
        caux << ENQ << '\x01';
	Getpack();
	switch (ptype)
	{
		      case TM:
				state=S1;
				break;
		      case RI:
				state=S;
				break;
		      case RR:
				state=SH;
				break;
		      case RF:
				state=SD;
				break;
		      case CN:
				state=C;
				break;
		      case NR:
				state=YAPP_START;
				break;
		      case TX:
				disppacket();
				break;
		      default:
				if (state==YAPP_ABORT)
				{
					yapp_message(ORG_STOP);
					break;
				}
				state=YAPP_ABORT;
				yapp_message(PROTOCOL_ERROR);
				break;
	}
}

/*
	Nel caso che non si fosse sentito (!!) riprova a mandare il SendInit
*/


static void SendInit_Retry (void)
{
	if(++sendinit_count > 6)
	{
		state=YAPP_ABORT;
		yapp_message(YAPP_TIMEOUT);
		return;
	}
	SendInit();
}

/*
	Qui viene trasmesso l'header e deciso sul da farsi successivo
*/

static void Sendhdr (void)
{
	int stlen;
	char temp[80];
	char nome[MAXPATH];
	char file[MAXFILE];
	char ext[MAXEXT];
	int i;

	xfercnt=0L;

	if (select_mode)
	{
		if(!primo_select_file)
		{
			if (lista_select_file->prossimo != NULL)
			{
			     time(&inizio);
			     altri_select_file=YES;
			     lista_select_file=lista_select_file->prossimo;
			     free(lista_select_file->precedente);
			}
			else
			{
				state=ST;
				return;
			}
		}
		sprintf(nome,"%s\\%s",lista_select_file->path,\
						lista_select_file->nome);
		strcpy(filename,lista_select_file->nome);
	}

	else
	{
		if (primo_multi_file)
			fnsplit(gruppo_file.ff_name,NULL,NULL,file,ext);
		else
		{
			if(!findnext(&gruppo_file))
			{
			   time(&inizio);
			   fnsplit(gruppo_file.ff_name,NULL,NULL,file,ext);
			   altri_multi_file=YES;
			}
			else
			{
				state=ST;
				return;
			}
		}
	fnmerge(nome,yapp_drive,yapp_path,file,ext);
	sprintf(filename,"%s%s",file,ext);
	}

	if((Yapp_File = fopen(nome,"rb")) == NULL)
	{
		fclose(Yapp_File);
		state=YAPP_ABORT;
		yapp_message(YAPP_IO_ERROR);
		return;
	}

	ltoa(filesize(Yapp_File),lunghezza,10);
	getftime(fileno(Yapp_File),&Yapp_filetime);
	sprintf(file_date,"%02u/%02u/%04u",
	Yapp_filetime.ft_day,Yapp_filetime.ft_month,Yapp_filetime.ft_year+1980);
	sprintf(file_time,"%02u:%02u:%02u",
	Yapp_filetime.ft_hour,Yapp_filetime.ft_min,Yapp_filetime.ft_tsec*2);

        sprintf(dataora,"%lx",Yapp_filetime);

        sprintf(accesso,"%x",(char)_chmod(nome,0));

	stlen=strlen(filename)+strlen(lunghezza)+strlen(dataora)+\
							strlen(accesso)+4;

        // Lo standard Yapp Header
        caux << SOH << (char) stlen << filename << NIL << lunghezza << NIL;

        // e l'extra.
        caux << dataora << NIL << accesso << NIL;

	showheader();
	yapp_message(S_HEADER);
	show_start_time();
	Getpack();
	switch(ptype)
	{
		case RF:
			yapp_message(T_FILE);
			state=SD;
			break;
		case CN:
			state=C;
			break;
		case NR:
			if ((Resume) && (*pkbuff=='R'))
			{
				for (i=2;i<pklen;i++)
					temp[i-2] = pkbuff[i];
				xfercnt=atol(temp);
				fseek(Yapp_File,xfercnt,SEEK_SET);
				if (xfercnt==filesize(Yapp_File))
				{
					state=SE;
					break;
				}
				yapp_message(FILE_RESUMED);
				state=SD;
			}
			else
			{
				yapp_message(DEST_STOP);
				state=YAPP_START;
			}
			break;
		case TX:
			disppacket();
			break;
		default:
			if (state==YAPP_ABORT)
			{
				yapp_message(ORG_STOP);
				break;
			}
			state=YAPP_ABORT;
			if (ptype==TM)
				yapp_message(YAPP_TIMEOUT);
			else
				yapp_message(PROTOCOL_ERROR);
			break;
	}
}

/*
	Qui vengono trasmessi i vari frame Yapp
*/


static void Senddata (void)
{
  register int i=0,cnt=0;
  char temp[258],chksum;
  int Ch;

  if (caux.isAny())
  {
	Getpack();
	if ((ptype == CN))
	{
		state = C;
		return;
	}
	else if ((ptype == TX))
			disppacket();
	     else
	     {
			state = YAPP_ABORT;
			return;
	     }
   }

   while (!feof(Yapp_File) && (cnt < YAPP_FRAME_WINDOW))
   {
        *(temp+cnt++) = (char) fgetc(Yapp_File);
        chksum += *(temp+cnt);
   }

   if (feof(Yapp_File)) cnt--;

   if (cnt)
   {
      caux << STX << (char) (cnt == 256 ? 0 : cnt );
      for (i = 0; i < cnt; i++)
                        caux << *(temp+i);
   }

   if (ptype == DT && file_resumed && resume_with_chksum)
                        caux << chksum;

    // Se sono stati trasmessi  meno di 256 caratteri allora il
    // trasferimento e' finito.

    if (cnt != 256)
               state = SE;

    xfercnt += (long) cnt;

    showbyte();
}

/*
	Invia sequenza EOT (Fine del trasferimento)
*/

static void SendEOT(void)
{
	yapp_message(T_STOP);
        caux << EOT << '\x01';
	Getpack();
	switch(ptype)
	{
		case AT:
			state=YAPP_START;
			break;
		case TX:
			disppacket();
			break;
		default:
			state=YAPP_START;
			break;
	}
}

// Invia sequenza Cancel per bloccare il trasferimento

static void Cancel(void)
{
  caux << CAN << NIL;
  delay (2000);
  state=CW;
}

/*
    Invia EOF (fine del file) e vede se ce ne sono altri da trasmettere
    altrimenti invia EOT
*/

static void SendEOF(void)
{
	fclose(Yapp_File);		// Chiudo il file !!!!
	yapp_message(FILE_T);
        caux << ETX << '\x01';
	Getpack();
	report_trasmit();
	switch(ptype)
	{
		case AF:
			if (select_mode)
			{
				if (primo_select_file || altri_select_file)
				{
				primo_select_file=NO;
				state = SH;
				}
			}
			else
			{
				if (primo_multi_file || altri_multi_file)
				{
				primo_multi_file=NO;
				state = SH;
				}
			}
			break;
		case TX:
			disppacket();
			break;
		default:
			state=YAPP_ABORT;
			if (ptype==TM) yapp_message(YAPP_TIMEOUT);
			else yapp_message(PROTOCOL_ERROR);
			break;
		}
}
/*
	Questa funzione comincia la Ricezione del trasferimento
*/


static void Receive (void)
{
	Getpack();
	switch (ptype)
	{
		case SI:
			yapp_message(R_FILE);
                        caux << ACK << '\x01';
			state=RH;
			break;
		case CN:
			state=C;
			break;
		case TX:
			disppacket();
			break;
		default:
			if (state==YAPP_ABORT)
			{
				yapp_message(DEST_STOP);
				break;
			}
			state=YAPP_ABORT;
			if(ptype==TM)
				yapp_message(YAPP_TIMEOUT);
			else
				yapp_message(PROTOCOL_ERROR);
			break;
		}
}



static void RcvHdr(void)
{
yapp_message(R_HEADER);
if (auto_start_download)
{
        caux << ACK << '\x01';   // RR
	auto_start_download=NO;
}
Getpack();
switch(ptype)
	{
	case HD:
		xferok=NO;
		xfercnt=0L;
		time(&inizio);
		split_header();
		struct ffblk ff;
		if(!findfirst(full_yapp_filename,&ff,0x07))
		{
			// Leggi informazioni sul file
			int len,date,att;

			len=(ff.ff_fsize==atol(lunghezza));

                        // Qui il confronto tra le date viene fatto confontando
                        // il contenuto della struttura tutta insieme (4byte=ulong).
                        // Lo so' non e' bello e sopratutto immediato ma
                        // cosi' si fa' prima.....

			if(*file_date)
				date=(*(unsigned long *)(&Yapp_filetime) == *(unsigned long *)(&ff.ff_ftime));
                        else
                                date=YES;

			if(*accesso)
				att=(ff.ff_attrib==Yapp_filemode);
                        else
                                att=YES;

			if(len)
			{
				if(date && att)
				{
                                        caux << NAK << (char) strlen(HaveThatFile) << HaveThatFile;
					state=YAPP_START;
					break;
				}
				else
				{
				// Se il file esiste con lo stesso nome ma con
				// diversi filedate o attributi..

					cambia_nome();

					if((Yapp_File = fopen(full_yapp_filename,"wb+")) == NULL)
					{
						fclose(Yapp_File);
						yapp_message(YAPP_IO_ERROR);
						delay(2000);
						state=YAPP_ABORT;
						break;
					}
                                        caux << ACK << '\x02';
                                        Yapp_window.printAt(2,30,ZaCc_Color.bin_text_att,"(Nome cambiato)");
				}
			}
			else if (ff.ff_fsize<atol(lunghezza)){
					if(date && att)
							if(Resume){
								if((Yapp_File = fopen(full_yapp_filename,"r+b")) == NULL)
								{
									fclose(Yapp_File);
									yapp_message(YAPP_IO_ERROR);
									delay(2000);
									state=YAPP_ABORT;
									break;
								}

								/*
								Qui viene richiesta la trasmissione della parte di file -256
								per evitare che eventuali porcherie residuate dal precedente trasferimento
								andato male rovinino il file in ricezione
								*/

                                                                ltoa(xfercnt=(filesize(Yapp_File)-256L),lunghezza_parziale,10);
								if (xfercnt <= 0L)
									xfercnt = 0L;
								fseek(Yapp_File,xfercnt,SEEK_SET);
								int stlen=(strlen(lunghezza_parziale)+3);
                                                                caux << NAK << (char) stlen << 'R' << NIL << lunghezza_parziale << NIL;
								yapp_message(FILE_RESUMED);

								// Serve per sapere cosa stampare alla fine
								file_resumed=YES;
								byte_file_resumed=filesize(Yapp_File);
                                                                Yapp_window.printAt(2,30,ZaCc_Color.bin_text_att,"(RESUME)");
								}
								else
								{
                                                                caux << NAK << (char) strlen(HaveThatFile) << HaveThatFile;
								state=YAPP_START;
								break;
								}
					else
					{
					// Se il file esiste con lo stesso nome ma con
					// diversi filedate o attributi..

					cambia_nome();

					if((Yapp_File = fopen(full_yapp_filename,"wb+")) == NULL)
					{
						fclose(Yapp_File);
						yapp_message(YAPP_IO_ERROR);
						delay(2000);
						state=YAPP_ABORT;
						break;
					}
                                        caux << ACK << '\x02';
					}
				}
			else
			{
			// Se il file esiste con lo stesso nome ma con
			// diversi filedate o attributi..

			cambia_nome();

			if((Yapp_File = fopen(full_yapp_filename,"wb+")) == NULL)
			{
				fclose(Yapp_File);
				yapp_message(YAPP_IO_ERROR);
				delay(2000);
				state=YAPP_ABORT;
				break;
			}
                        caux << ACK << '\x02';
			}
		}
		else
		{
			// Se il file non esiste allora creane uno di nuovo

			if((Yapp_File = fopen(full_yapp_filename,"wb+")) == NULL)
			{
				fclose(Yapp_File);
				yapp_message(YAPP_IO_ERROR);
				delay(2000);
				state=YAPP_ABORT;
				break;
			}
                        caux << ACK << '\x02';
		}

		showheader();
		show_start_time();
		state=RD;
		break;
	case SI:
		state=RH;
		break;
	case CN:
		state=C;
		break;
	case ET:
                caux << ACK << '\x04';
		state=YAPP_START;
		break;
	case TX:
		disppacket();
		break;
	default:
		if (state==YAPP_ABORT)
		{
			yapp_message(DEST_STOP);
			break;
		}
		state=YAPP_ABORT;
                if (ptype==TM)
                        yapp_message(YAPP_TIMEOUT);
                else
                        yapp_message(PROTOCOL_ERROR);
	}
}

/*
	Riceve i frame Yapp e tratta eventuali errori
*/

static void RcvData (void)
{
	yapp_message(R_FILE);
	Getpack();
	switch(ptype){
		case DT:
			fwrite(pkbuff,pklen,1,Yapp_File);
			xfercnt += (long) pklen;
			showbyte();
			state=RD;
			break;
		case EF:
			fclose(Yapp_File);
			change_date_attr();
			xferok=OK;
			yapp_message(FILE_R);
                        caux << ACK << '\x03';
			report_receive();
			state=RH;
			break;
		case CN:
			yapp_message(ORG_STOP);
			state=C;
			break;
		case TX:
			disppacket();
			break;
		default:
                        if (state==YAPP_ABORT)
                        {
				yapp_message(DEST_STOP);
				break;
			}
			state=YAPP_ABORT;
                        if(ptype==TM)
                                yapp_message(YAPP_TIMEOUT);
                        else
                                yapp_message(PROTOCOL_ERROR);
			break;
		}
}


static void CanWait(void)
{
	Getpack();
	switch(ptype)
	{
		case CA:
			state=YAPP_START;
			break;
		case CN:
                        caux << ACK << '\x05';
			break;
		case TM:
		case UK:
			state=YAPP_START;
			break;
		case TX:
			disppacket();
			break;
		default:
			break;
	}
}


static void CanRecd (void)
{
  yapp_message(ORG_STOP);
  caux << ACK << '\x05';
  delay(3000);
  state=YAPP_START;
}



static void yapp_message(int num)
{
   char *msg[] = {
		   "TNC in transparent mode               ",
		   "TNC in converse mode                  ",
		   "Trasferimento iniziato                ",
		   "Trasferimento concluso                ",
		   "Trasmissione intestazione             ",
		   "Ricezione intestazione                ",
		   "Trasmissione file                     ",
		   "Ricezione file                        ",
		   "File trasmesso                        ",
		   "File ricevuto                         ",
		   "Trasferimento bloccato a destinazione ",
		   "Trasferimento bloccato in origine     ",
		   " *** Timeout!                         ",
		   " *** Errore di protocollo             ",
		   "Errore di accesso al disco            ",
		   "RESUME                                "
		  };
      Yapp_window.printAt(15,15,ZaCc_Color.bin_text_att,msg[num]);
}

static void showstate (void)
{
  Yapp_window.printAt(14,15,ZaCc_Color.bin_text_att,state_desc[state]);
}

static void showbyte (void)
{
  char temp[20];
  long ore=0L,minuti=0L,secondi=0L;
  ldiv_t d;
  time_t parziale=0L;
  sprintf(temp,"%-6lu",xfercnt);
  Yapp_window.printAt(10,21,ZaCc_Color.bin_text_att,temp);
  time(&parziale);
  if ((differenza_tempo=(time_t)difftime(parziale,inizio)) > 0L)
  {
	  sprintf(cps,"%-6lu",(xfercnt/differenza_tempo));
          Yapp_window.printAt(10,39,ZaCc_Color.bin_text_att,cps);
	  d = ldiv(differenza_tempo,3600L);
	  ore=d.quot;
	  d = ldiv(d.rem,60L);
	  minuti=d.quot; secondi=d.rem;
  }
  sprintf(tempo_parziale,"%02ld:%02ld:%02ld",ore,minuti,secondi);
  Yapp_window.printAt(12,20,ZaCc_Color.bin_text_att,tempo_parziale);
}


static void showheader (void)
{
  Yapp_window.printAt(2,14,ZaCc_Color.bin_text_att,filename);
  Yapp_window.printAt(3,14,ZaCc_Color.bin_text_att,lunghezza);
  if (*file_date)
  {
        Yapp_window.printAt(4,14,ZaCc_Color.bin_text_att,file_date);
        Yapp_window.printAt(5,14,ZaCc_Color.bin_text_att,file_time);
  }
}

static void show_start_time(void)
{
  tm *ora;
  char temp[20];
  ora=localtime(&inizio);
  sprintf(temp,"%02i:%02i:%02i",ora->tm_hour,ora->tm_min,ora->tm_sec);
  Yapp_window.printAt(11,20,ZaCc_Color.bin_text_att,temp);
}

/*
        Visualizza il msg pervenuto dal corrispondente
*/

static void disppacket (void)
{
  Term_window.putc('\n');
  for (int i=0;i<pklen;i++)
                Term_window.putc(pkbuff[i]);
}
/*
        Invia un msg al corrispondente
*/

static void SendMSG (char *buf)
{
  caux << DLE << (char) strlen(buf) << buf;
}

static void split_header(void)
{
  char temp[80];
  char file[MAXFILE];
  char ext[MAXEXT];
  time_t t;
  int i=0,j=0;

  // Preleva il nome dall'header

  do { temp[i]=pkbuff[i]; } while(pkbuff[i++]);

  // e tiene solo il nome+ext

  if(!simple_download)
  {
	fnsplit(temp,NULL,NULL,file,ext);
	sprintf(filename,"%s%s",file,ext);
	fnmerge(full_yapp_filename,ZaCc_Var.download_drive,\
					ZaCc_Var.download_dir,file,ext);
  }
  else
  {
	fnsplit(full_yapp_filename,NULL,NULL,file,ext);
	sprintf(filename,"%s%s",file,ext);
	fnmerge(full_yapp_filename,ZaCc_Var.download_drive,\
					ZaCc_Var.download_dir,file,ext);
  }

  // Preleva la lunghezza..

  while(pkbuff[i])
        lunghezza[j++]=pkbuff[i++];

  // Se non c'e' piu' niente ritorna

  if (++i == pklen) return;

  // Altrimenti tenta di prelevare la data e l'ora del file

  j=0;

  while(pkbuff[i])
        temp[j++]=pkbuff[i++];
  temp[j]=NULL;

  if(sscanf(temp,"%lx",&Yapp_filetime)){
	sprintf(file_date,"%02u/%02u/%04u",Yapp_filetime.ft_day,
			Yapp_filetime.ft_month,Yapp_filetime.ft_year+1980);

	sprintf(file_time,"%02u:%02u:%02u",Yapp_filetime.ft_hour,
				Yapp_filetime.ft_min,Yapp_filetime.ft_tsec*2);
	}

  if (++i == pklen) return;
  j=0;

  // Stesso discorso con gli attributi.

  while(pkbuff[i])
                temp[j++]=pkbuff[i++];
  temp[j]=NULL;

  if(sscanf(temp,"%x",&Yapp_filemode))
			strcpy(accesso,temp);
}

/*
 Mette il TNC in Transaparent mode da Command Mode solo se Tnc_type > 1
 Nel caso con Tnc_type = 0 (usato in prova tra due computer) ritorna
*/

static void Trans (void)
{
if (Tnc_type) caux << CR << 'T' << CR;
}

/*
 Mette il TNC in Converse mode da Command Mode solo se Tnc_type > 1
 Nel caso con Tnc_type = 0 (usato in prova tra due computer) ritorna
*/

static void Conv (void)
{
if (Tnc_type) caux << CR << 'K' << CR;
}

/*
  Cambia file date e time di creazione e attributi solo se sono stati ricevuti
  dal corrispondente.
*/

static void change_date_attr(void)
{
	if(*file_date)
	{
		Yapp_File=fopen(full_yapp_filename,"r+");
		setftime(fileno(Yapp_File),&Yapp_filetime);
		fclose(Yapp_File);
	}
	if(*accesso)
		_chmod(full_yapp_filename,1,(int)Yapp_filemode);
}

/*
  Cambia nome (estensione) al file in ricezione
*/

static void cambia_nome (void)
{
    int i=0;
    char my_ext[MAXEXT];
    char drive[MAXDRIVE];
    char dir[MAXDIR];
    char file[MAXFILE];
    char ext[MAXEXT];
    struct ffblk ff;

    // prova tutte le estensioni da .000 a .999 finche' non ne trova una
    // che non esiste gia'..
    // Di solito basta un solo tentativo (.000)

    do
    {
        fnsplit(full_yapp_filename,drive,dir,file,ext);
        sprintf(my_ext,".%003i",i++);
        fnmerge(full_yapp_filename,drive,dir,file,my_ext);
    }
    while(!findfirst(full_yapp_filename,&ff,0));
    sprintf(filename,"%s%s",file,my_ext);
}

